/**
 * File:    Segway.cpp
 *
 * Author:  David Zemon
 * Project: Segway
 */

// Includes
#include "Segway.h"
#include <stdio.h>
static uint32_t g_cog_stack[THREAD_STACK_SIZE];
static _thread_state_t g_thread_data;

// Main function
void main (void) {
	uint32_t loopTimer; // Continuous 20ms timer initialized at the beginning of each loop

	int16_t leftOut = 0;	// PWM value (0-255) output to the left wheel driver
	int16_t rightOut = 0;// PWM value (0-255) output to the right wheel driver
	PWMDualMotors motors;	// Motor pins
	uint32_t backendCogID; // Cog ID of the backend cog (constantly updating global var for leanAngle)

	setup(&motors, &backendCogID);

	while (1) {
		loopTimer = CNT + LOOP_CNT_MILLIS * MILLISECOND;

		// If another thread through an error, stop the motors and hault execution
		if (g_hardFault) {
			PWMDualMotorsDrive(&motors, 0, 0);
			while(1);
		}

		assignOutputs(g_leanAngle, g_turnAngle, &leftOut, &rightOut);
		PWMDualMotorsDrive(&motors, leftOut, rightOut);
#if (defined DEBUG && defined DEBUG_VERBOSE)
		printf("\t%i\t%i\t%i\t%i\n", g_leanAngle, g_turnAngle, leftOut,
				rightOut);
#endif

		// Wait for the 20ms time period to finish
		waitcnt(loopTimer);
	}
}

void setup (PWMDualMotors *motors, uint32_t *backendCogID) {
#if (defined DEBUG && defined DEBUG_VERBOSE)
	printf("Hello! Starting Segway!\n");
#endif

#ifdef DEBUG
	GPIODirModeSet(DEBUG_LED, GPIO_DIR_OUT); //debugging tool for finding length of one loop
#endif
	// Initialize the motors
	PWMDualMotorsInit(motors, LEFT_MOTOR_PIN, RIGHT_MOTOR_PIN);

	// Initialize lean angle read cog
	*backendCogID = _start_cog_thread(g_cog_stack, segway_backend, NULL,
			&g_thread_data);
}

uint8_t assignOutputs (const int8_t leanAngle, const int8_t turnAngle,
		int16_t *leftOut, int16_t *rightOut) {
	// Calculate lean angle only if applicable (outside NULL zone)
	if (!(abs(leanAngle) < LEAN_NULL_ZONE)) {
		// use MAX_LEAN or -MAX_LEAN if leanAngle is greater than maximum angle
		// output proportional to leanAngle
		float usefulLean;
		if (leanAngle >= MAX_LEAN)
			usefulLean = MAX_LEAN;
		else if (leanAngle <= -MAX_LEAN)
			usefulLean = -MAX_LEAN;
		else
			usefulLean = leanAngle;

		// Create an output for each motor
		*leftOut = *rightOut = runPID(usefulLean);
	}

	// Add or subtract up to 20 from each motor to compensate for turning, bringing
	// the total range to -500 to 500
//	*leftOut += turnAngle * (float) MAX_DIFFERENTIAL / MAX_TURN_POWER / TURN_DIVISOR;
//	*rightOut -= turnAngle * (float) MAX_DIFFERENTIAL / MAX_TURN_POWER / TURN_DIVISOR;

	return 0;
}

int16_t runPID (int8_t leanAngle) {
	static float errorPrev = 0;
	static float integral = 0;
	float derivative;
	float error;

	// Proportional
	error = KP * leanAngle;

	// Derivative
	derivative = error - errorPrev;

	// Integral
	// Do we need to reset the integral because we're now correcting in the opposite direction?
	if ((errorPrev > 0 && 0 > error) || (error > 0 && 0 > errorPrev))
		integral = 0;
	else
		integral += error;

	errorPrev = error;

	return error + KD * derivative + KI * integral;
}
